/******************************************************************************
 *
 *  NETWORK CLASS
 *
 ******************************************************************************/
class Network
{
  /****************************************************************************
   *
   *  PROPERTIES
   *
   ****************************************************************************/

  // TEMPORARY NEURON ARRAY
  Neuron [] hiddenLayer;

  // INPUT NEURONS
  Neuron [] inputLayerNeurons; 

  // HIDDEN NEURONS (MULTIPLE LAYERS)
  ArrayList<Neuron[]> hiddenLayerNeurons = new ArrayList();

  // OUTPUT NEURONS
  Neuron [] outputLayerNeurons;

  int lastResponse = -1;


  /****************************************************************************
   *
   *  CONSTRUCTOR:
   *
   *  CREATE A NETWORK SPECIFYING NUMBERS OF INPUTS AND OUTPUTS
   *  AND NUMBER OF OUTPUTS, E.G. NETWORK(4,3)
   *
   ****************************************************************************/
  Network(int _inputs, int _outputs)
  {
    // INITIALIZE INPUT NEURONS
    inputLayerNeurons   = new Neuron [_inputs];

    // INITIALIZE HIDDEN NEURONS
    for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++) hiddenLayerNeurons.add(new Neuron [NR_OF_HIDDEN_NEURONS.get(i)]);

    // INITIALIZE OUTPUT NEURONS
    outputLayerNeurons  = new Neuron [_outputs];

    // SET UP THE NETWORK TOPOLOGY
    for (int i = 0; i < inputLayerNeurons.length; i++) inputLayerNeurons[i] = new Neuron();

    if (NR_OF_HIDDEN_LAYERS == 0)
    { 
      // NO HIDDEN LAYERS: INPUT CONNECTS DIRECTLY TO OUTPUT
      for (int k = 0; k < outputLayerNeurons.length; k++) outputLayerNeurons[k] = new Neuron(inputLayerNeurons);
    } else
    {
      // CONNECT THE HIDDEN LAYERS
      for (int m=0; m<NR_OF_HIDDEN_LAYERS; m++)
      {   
        hiddenLayer = hiddenLayerNeurons.get(m);
        if (m == 0)
          // FIRST HIDDEN LAYER: CONNECT TO INPUT LAYER
          for (int j = 0; j < hiddenLayer.length; j++) hiddenLayer[j] = new Neuron(inputLayerNeurons);
        if (m > 0)
          // MIDDLE HIDDEN LAYER: CONNECT TO THE PREVIOUS HIDDEN LAYER
          for (int j = 0; j < hiddenLayer.length; j++) hiddenLayer[j] = new Neuron(hiddenLayerNeurons.get(m-1));
        if (m == NR_OF_HIDDEN_LAYERS-1)
          // LAST HIDDEN LAYER: CONNECT TO OUTPUT LAYER
          for (int k = 0; k < outputLayerNeurons.length; k++)
            outputLayerNeurons[k] = new Neuron(hiddenLayer);
        // UPDATE LAYER
        hiddenLayerNeurons.set(m, hiddenLayer);
      } // for (int m=0; m<NR_OF_HIDDEN_LAYERS; m++)
    } // if (NR_OF_HIDDEN_LAYERS == 0)
  } // Network


  /****************************************************************************
   *
   *  CALCULATE THE OUTPUTS FOR A TESTING OR TRAINING IMAGE (inputs)
   *
   ****************************************************************************/
  int respond(float [] inputs)
  {
    float [] responses = new float [outputLayerNeurons.length];

    // FEED FORWARD
    // SIMPLY SET THE INPUT LAYER TO DISPLAY THE INPUTS
    for (int i = 0; i < inputLayerNeurons.length; i++) inputLayerNeurons[i].neuronOutput = inputs[i];

    // FEED FORWARD THE HIDDEN LAYER(S)
    for (int j = 0; j < NR_OF_HIDDEN_LAYERS; j++)
    {
      hiddenLayer = hiddenLayerNeurons.get(j);
      for (int n = 0; n < hiddenLayer.length; n++) hiddenLayer[n].respond();
    }

    // AND FINALLY FEED FORWARD TO THE OUTPUT LAYER
    for (int k = 0; k < outputLayerNeurons.length; k++) responses[k] = outputLayerNeurons[k].respond();

    // NOW CHECK THE BEST RESPONSE:
    int response = -1;
    float best = max(responses);
    for (int a = 0; a < responses.length; a++)
      if (responses[a] == best) response = a;

    fLastAccuracy = (best+1) * 50;

    lastResponse = response;

    return response;
  } // respond()


  /****************************************************************************
   *
   *  TRAIN THE NETWORK
   *
   ****************************************************************************/
  void train(float [] outputs)
  {
    // ADJUST THE OUTPUT LAYER
    for (int k = 0; k < outputLayerNeurons.length; k++)
    {
      outputLayerNeurons[k].finderror(outputs[k]);
      outputLayerNeurons[k].train();
    }

    // PROPAGATE BACK TO HIDDEN LAYERS
    for (int j = 0; j < NR_OF_HIDDEN_LAYERS; j++)
    {
      hiddenLayer = hiddenLayerNeurons.get(j);
      for (int n = 0; n < hiddenLayer.length; n++) hiddenLayer[n].train();
    }

    // THE INPUT LAYER DOESN'T LEARN:
    // IT IS SIMPLY THE INPUTS
  } // train()


  /****************************************************************************
   *
   *  DRAW THE NETWORK
   *
   ****************************************************************************/
  void draw()
  {
    // NOTE, THIS DRAW IS HARD-CODED FOR NETWORK(196,49,10)
    // WHICH REFLECTS MY USE OF THE MNIST DATABASE OF HANDWRITTEN DIGITS

    pushMatrix();

    // OUTPUT LAYER
    // THIS IS SLIGHTLY TRICKY -- I'VE SWITCHED THE ORDER SO THE OUTPUT
    // NEURONS ARE ARRANGE 1,2,3...8,9,0 RATHER THAN 0,1,2...7,8,9
    // (THAT'S WHAT THE (K+9) % 10 IS DOING)
    for (int k = 0; k < outputLayerNeurons.length; k++) {
      pushMatrix();
      translate(((k+9)%10) * 600 / 20.0 + output_offset_x, 20);
      outputLayerNeurons[k].draw("o", k);
      popMatrix();
    }

    // HIDDEN LAYER(S)
    for (int l=0; l<NR_OF_HIDDEN_LAYERS; l++)
    {
      hiddenLayer = hiddenLayerNeurons.get(l);  
      for (int m = 0; m < hiddenLayer.length; m++)
      {
        pushMatrix();
        translate((m % HIDDEN_COLUMNS[l]) * 600 / 30.0 + hidden_offset_x[l], (m / HIDDEN_COLUMNS[l]) * 600 / 30.0 + hidden_offset_y[l]);
        hiddenLayer[m].draw("h", m);
        popMatrix();
      }
    }

    // INPUT LAYER
    for (int i = 0; i < inputLayerNeurons.length; i++)
    {
      pushMatrix();
      translate((i%INPUT_COLUMNS) * 600 / 30.0 + input_offset_x, (i/INPUT_COLUMNS) * 600 / 30.0 + input_offset_y);
      inputLayerNeurons[i].draw("i", i);
      popMatrix();
    }

    popMatrix();
  } // draw()
} // Network